/*
tri2map - raw triangle to Quake .map terrain convertor

Copyright (C) 2000  Seth Galbraith

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


Seth Galbraith
"The Serpent Lord"
http://www.planetquake.com/simitar
sgalbrai@krl.org
*/

#include <stdio.h>
#include <stdlib.h>

// pOx - additions
#include <console.h>
#include <SIOUX.h>
#include "cmdlib.h"
#include "mac_io.h"
#include "mathlib.h"
#include "objlib.h"
#include "trilib.h"
#include "l3dslib.h"
#include "Matrix.h"

#define TRITYPE_RAW 1
#define TRITYPE_CAD 2
#define TRITYPE_3DS 3
#define TRITYPE_OBJ 4

qboolean	verbose, superverbose; // Not used, just for objlib's sake

qboolean	flipnormals = false; // Flip normals (for troublesome 3DS files)

int write_terrain_triangle (FILE *f, float flr, float quark,
	float ax, float ay, float az,
	float bx, float by, float bz,
	float cx, float cy, float cz)
{
	if (az <= flr && bz <= flr && cz <= flr) return 1;

	fprintf(f, " {\n");

	if (bz > flr || cz > flr)
		fprintf(f, "  ( %f %f %f ) ( %f %f %f ) ( %f %f %f ) rock1_2 0 0 0 1 1\n",
			cx, cy, cz,   bx, by, bz,    bx, by, bz - 128); // SIDE BC
	if (bz > flr || az > flr)
		fprintf(f, "  ( %f %f %f ) ( %f %f %f ) ( %f %f %f ) rock1_2 0 0 0 1 1\n",
			bx, by, bz,   ax, ay, az,    ax, ay, az - 128); // SIDE AB
	if (az > flr || cz > flr)
		fprintf(f, "  ( %f %f %f ) ( %f %f %f ) ( %f %f %f ) rock1_2 0 0 0 1 1\n",
			ax, ay, az,   cx, cy, cz,    cx, cy, cz - 128); // SIDE AC

	fprintf(f, "  ( %f %f %f ) ( %f %f %f ) ( %f %f %f ) black 0 0 0 1 1\n",
		0.0f,0.0f,flr, 4096.0f,0.0f,flr, 0.0f,4096.0f,flr); // BOTTOM

	if (quark)
	{
		bx = (bx - ax) * 128.0f / quark + ax;
		by = (by - ay) * 128.0f / quark + ay;
		bz = (bz - az) * 128.0f / quark + az;
		cx = (cx - ax) * 128.0f / quark + ax;
		cy = (cy - ay) * 128.0f / quark + ay;
		cz = (cz - az) * 128.0f / quark + az;
		fprintf(f, "  ( %f %f %f ) ( %f %f %f ) ( %f %f %f ) ground1_1 0 0 0 1 1 //TX1\n",
			ax, ay, az,   bx, by, bz,   cx, cy, cz   ); // TOP
	}
	else fprintf(f, "  ( %f %f %f ) ( %f %f %f ) ( %f %f %f ) ground1_1 0 0 0 1 1\n",
			ax, ay, az,   bx, by, bz,   cx, cy, cz   ); // TOP

	fprintf(f, " }\n");

	return 0;
}

// pOx - makes supporting more model formats a little neater (if they fail)
void write_terrain_header (FILE *o)
{
	fprintf (o, "{\n");
	fprintf (o, " \"classname\" \"worldspawn\"\n");
	fprintf (o, " \"message\" \"No Name\"\n");
	fprintf (o, " \"worldtype\" \"0\"\n");
}

int convert (char *inname, char *out, float hscale, float vscale,
	float xoff, float yoff, float zoff, float floor, float quark)
{
	FILE *i, *o;
	int length, type = 0, j;
	char str [256], in[1024];
	float ax,ay,az, bx,by,bz, cx,cy,cz;

	printf ("Input  File Name: %s\n", inname);
	printf ("Output File Name: %s\n\n", out);
	
	inname = strlower(inname);// Don't choke on uppercase extensions
	strcpy (in, inname);
	
	// first check if the extension is specified
	length = strlen (in);
	if (length < 5) type = 0;
	else if ( !strcmp(in+(length-4), ".raw") ) type = TRITYPE_RAW;
	else if ( !strcmp(in+(length-4), ".cad") ) type = TRITYPE_CAD;
	else if ( !strcmp(in+(length-4), ".3ds") ) type = TRITYPE_3DS;
	else if ( !strcmp(in+(length-4), ".obj") ) type = TRITYPE_OBJ;
	
	// If not found, check for each file type
	if (!type)
	{
		StripExtension (in);
		strcat (in, ".3ds");
		
		if (Sys_FileTime(in) > 0)
			type = TRITYPE_3DS;
		else
		{
			StripExtension (in);
			strcat (in, ".obj");
			
			if (Sys_FileTime(in) > 0)
				type = TRITYPE_OBJ;
			else
			{
				StripExtension (in);
				strcat (in, ".raw");
			
				if (Sys_FileTime(in) > 0)
					type = TRITYPE_RAW;
				else
				{
					StripExtension (in);
					strcat (in, ".cad");
			
					if (Sys_FileTime(in) > 0)
						type = TRITYPE_CAD;	
				}
			}
		}
	}
	
	if (type == 0)
	{
		printf ("ERROR: input file type should be .raw, .cad, .3ds or .obj\n");
		return 1;
	}
	
	// 3DS and OBJ have their own open routines
	if (type > TRITYPE_CAD)
		i = Sys_fopen (in,  "ra");
	
	// Announce the file we're using since the auto-detect can overide files
	printf ("Parsing %s...\n", in);
	
	if (type == TRITYPE_3DS)
	{
		int			tottris;
		triangle_t	*triptr;
	
		Load3DSTriangleList (in, &triptr, &tottris);
		
		o = Sys_fopen (out, "wa");
    	write_terrain_header (o);
		
		for (j=0 ; j<tottris ; j++)
		{
			// default is reversed cause that's what C4DXL spits out...
			if (!flipnormals)
			{
				cx = triptr[j].verts[0][0]; ax = triptr[j].verts[2][0];
				cy = triptr[j].verts[0][1]; ay = triptr[j].verts[2][1];
				cz = triptr[j].verts[0][2]; az = triptr[j].verts[2][2];
			}
			else
			{
				ax = triptr[j].verts[0][0]; cx = triptr[j].verts[2][0];
				ay = triptr[j].verts[0][1]; cy = triptr[j].verts[2][1];
				az = triptr[j].verts[0][2]; cz = triptr[j].verts[2][2];
			}
	
			bx = triptr[j].verts[1][0];
			by = triptr[j].verts[1][1];
			bz = triptr[j].verts[1][2];
			
			write_terrain_triangle (o, floor, quark,
				ax * hscale + xoff, ay * hscale + yoff, az * vscale + zoff,
				bx * hscale + xoff, by * hscale + yoff, bz * vscale + zoff,
				cx * hscale + xoff, cy * hscale + yoff, cz * vscale + zoff);
		}
		
		free(triptr);
	}
	else if (type == TRITYPE_OBJ)
	{
		OBJmodel_t	*model;
		OBJgroup_t	*group;
		vec3_t		trans = {90.0, 0.0, 0.0};
		
		model = LoadOBJModel (in, OBJ_TRIANGLES, -1, trans);
    	
    	if (!model || !model->vertices || model->numvertices < 3 || !model->polys)
    	{
    		printf ("ERROR: No valid geometry found in OBJ model.\n");
			return 1;
    	}
    	
    	o = Sys_fopen (out, "wa");
    	write_terrain_header (o);
    	
		group = model->groups;
				
		while(group)
		{
			for (j = 0; j < group->numpolys; j++)
			{
				// We asked for tris, so that's all we worry about...
				ax = model->vertices[3*T(group->polys[j]).vindices[0]+X];
				ay = model->vertices[3*T(group->polys[j]).vindices[0]+Y];
				az = model->vertices[3*T(group->polys[j]).vindices[0]+Z];	
				bx = model->vertices[3*T(group->polys[j]).vindices[1]+X];
				by = model->vertices[3*T(group->polys[j]).vindices[1]+Y];
				bz = model->vertices[3*T(group->polys[j]).vindices[1]+Z];
				cx = model->vertices[3*T(group->polys[j]).vindices[2]+X];
				cy = model->vertices[3*T(group->polys[j]).vindices[2]+Y];
				cz = model->vertices[3*T(group->polys[j]).vindices[2]+Z];
				
				write_terrain_triangle (o, floor, quark,
					ax * hscale + xoff, ay * hscale + yoff, az * vscale + zoff,
					bx * hscale + xoff, by * hscale + yoff, bz * vscale + zoff,
					cx * hscale + xoff, cy * hscale + yoff, cz * vscale + zoff);
			}
			
			group = group->next;	
		}
		
		FreeOBJModel (model);
	}
	else
	{
		o = Sys_fopen (out, "wa");
    	write_terrain_header (o);
		
		while (1)
		{
			if (Q_fgets (str, 255, i) == NULL) break;
			if (type == 1)
			{
				if (strlen (str) < 5) continue;
				sscanf (str, "%f %f %f %f %f %f %f %f %f",
					&ax,&az,&ay, &bx,&bz,&by, &cx,&cz,&cy);
				ax *= -1;  bx *= -1;  cx *= -1;
				ay *= -1;  by *= -1;  cy *= -1;
			}
			if (type == 2)
			{
				if (str[0] != 'F' || str[1] != ' ') continue;
				sscanf (str, "F %f %f %f %f %f %f %f %f %f",
					&ay,&az,&ax, &by,&bz,&bx, &cy,&cz,&cx);
				ay *= -1;  by *= -1;  cy *= -1;
			}
			write_terrain_triangle (o, floor, quark,
				ax * hscale + xoff, ay * hscale + yoff, az * vscale + zoff,
				bx * hscale + xoff, by * hscale + yoff, bz * vscale + zoff,
				cx * hscale + xoff, cy * hscale + yoff, cz * vscale + zoff);
		}
	}

	fprintf (o, "}\n");
	fprintf (o, "{\n");
	fprintf (o, " \"classname\" \"info_player_start\"\n");
	fprintf (o, " \"origin\" \"0 0 0\"\n");
	fprintf (o, " \"angle\" \"360\"\n");
	fprintf (o, "}\n");

	if (type > TRITYPE_CAD)// 3DS or OBJ
		Sys_fclose (i);
		
	Sys_fclose (o);

	return 0;
}

int help ()
{
	printf ("USAGE: tri2map [options] <sourcefile> [destfile]\n");
	printf ("OPTIONS:\n");
	printf ("        -fn         : flip input model's normals (3DS only)\n");
	printf ("        -h [float]  : horizontal scale (default = 64)\n");
	printf ("        -v [float]  : vertical   scale (default = 64)\n");
	printf ("        -s [float]  : sets both horizontal and vertical scale (default = 64)\n");
	printf ("        -x [float]  : X offset (lattitude) in Quake units (default = 0)\n");
	printf ("        -y [float]  : Y offset (longitude) in Quake units (default = 0)\n");
	printf ("        -z [float]  : Z offset (altitude)  in Quake units (default = 0)\n");
	printf ("        -f [float]  : height of brush bottoms in Quake units (default = -4096)\n");
	printf ("        -q [float]  : use Quake Army Knife (QuArK)'s \"three points\"\n");
	printf ("                      texture alignment system to fit textures on faces\n");
	printf ("                      this float value is the number of texture pixels\n");
	printf ("                      to stretch along each edge of the triangles\n");

	return 1;
}

int main (int argc, char ** argv)
{
	int i;

	char  outfile [100];
	float hscale = 64;
	float vscale = 64;
	float xoff   = 0;
	float yoff   = 0;
	float zoff   = 0;
	float floor  = -4096;
	float quark  = 0;
	
	SIOUXSettings.userwindowtitle  = "\potri2map console log";
	SIOUXSettings.columns          = 104;
	SIOUXSettings.initializeTB     = true;
	SIOUXSettings.standalone       = true;
	SIOUXSettings.setupmenus       = true;
	SIOUXSettings.autocloseonquit  = false;
	SIOUXSettings.asktosaveonclose = false;
	SIOUXSettings.showstatusline   = false;
	
	argc = ccommand(&argv);// pOx - get the command line
	
	printf ("tri2map - raw triangle to Quake .map terrain convertor\n");
	printf ("Copyright (C) 2000  Seth Galbraith\n");
	printf ("\nPorted to MacOS by Frank Condello <pox@planetquake.com>\n\n");

	if (argc < 2) {help (); return 1;}
	
	for (i=1; i < argc; i++)
	{
		outfile[0] = '\0';
		
		if (argv[i][0] != '-')
			break;
		else if (!strcmp (argv[i],"-fn"))
			flipnormals = true;
		else if (!strcmp (argv[i], "-h"))
		{
			hscale = atof (argv[i + 1]);
			i++;		
		}
		else if (!strcmp (argv[i], "-v"))
		{
			vscale = atof (argv[i + 1]);
			i++;
		}
		else if (!strcmp (argv[i], "-s"))
		{
			hscale = atof (argv[i + 1]);
			vscale = hscale;
			i++;
		}
		else if (!strcmp (argv[i], "-x"))
		{
			xoff   = atof (argv[i + 1]);
			i++;
		}
		else if (!strcmp (argv[i], "-y"))
		{
			yoff   = atof (argv[i + 1]);
			i++;
		}
		else if (!strcmp (argv[i], "-z"))
		{
			zoff   = atof (argv[i + 1]);
			i++;
		}
		else if (!strcmp (argv[i], "-f"))
		{
			floor  = atof (argv[i + 1]);
			i++;
		}
		else if (!strcmp (argv[i], "-q"))
		{
			quark  = atof (argv[i + 1]);
			i++;
		}
		else
		{
			printf ("ERROR: unrecognized option '%s'.\n", argv[i]);
			help ();
			return 1;
		}
	}
	
	if (i != argc - 2)
	{
		strcpy (outfile, argv[i]);
		StripExtension (outfile);
		strcat (outfile, ".map");
	}
	else
		strcpy (outfile, argv[i+1]);
	
	convert (argv[i], outfile, hscale, vscale, xoff, yoff, zoff, floor, quark);
	
	return 0;
}
